# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
"""Markdown related utilities for Sphinx DocFX YAML extension."""

import os
import re
import shutil
from collections.abc import MutableSet
from pathlib import Path
from typing import Iterable

import sphinx.application
from docuploader import shell


def _reformat_codeblocks(content: str) -> str:
    """Formats codeblocks from ``` to <pre>.

    Args:
        content (str): The string to reformat.

    Returns:
        str: The reformatted string.
    """
    triple_backtick = "```"
    current_tag = "<pre>"
    next_tag = "</pre>"
    # If there are no proper pairs of triple backticks, don't format docstring.
    if content.count(triple_backtick) % 2 != 0:
        print(
            f"Docstring is not formatted well, missing proper pairs of triple backticks (```): {content}"
        )
        return content
    while triple_backtick in content:
        content = content.replace(triple_backtick, current_tag, 1)
        # Alternate between replacing with <pre> and </pre>.
        current_tag, next_tag = next_tag, current_tag

    return content


def _reformat_code(content: str) -> str:
    """Formats code from ` to <code>.

    Args:
        content (str): The string to reformat.

    Returns:
        str: The reformatted string.
    """
    reformatted_lines = []

    code_pattern = "`[^`\n]+`"
    code_start = "<code>"
    code_end = "</code>"
    prev_start = prev_end = 0
    # Convert `text` to <code>text</code>
    for matched_obj in re.finditer(code_pattern, content):
        start = matched_obj.start()
        end = matched_obj.end()
        code_content = content[start + 1 : end - 1]

        reformatted_lines.append(content[prev_end:start])
        reformatted_lines.append(f"{code_start}{code_content}{code_end}")
        prev_start, prev_end = start, end

    reformatted_lines.append(content[prev_end:])

    return "".join(reformatted_lines)


def reformat_markdown_to_html(content: str) -> str:
    """Applies changes from markdown syntax to equivalent HTML.

    Acts as a wrapper function to format all Markdown to HTML.

    Markdown syntax cannot be used within HTML elements, and must be converted
    at YAML level.

    Args:
        content: the string to be reformatted.

    Returns:
        Content that has been formatted with proper HTML.
    """

    content = _reformat_codeblocks(content)
    content = _reformat_code(content)

    return content


def _parse_markdown_header(current_line: str, prev_line: str) -> str:
    """Parses the H1 markdown header if found.

    Args:
        current_line: line of markdown text to inspect.
        prev_line: previous line to use if we found line divider for H1.

    Returns:
        Header for the markdown file if valid header is found.
    """
    # Markdown h1 prefix should have only 1 of '#' character followed by exactly one space.
    h1_header_prefix = "# "
    if h1_header_prefix in current_line and current_line.count("#") == 1:
        # Check for proper h1 header formatting, ensure there's more than just
        # the hashtag character, and exactly only one space after the hashtag.
        if (
            not current_line[current_line.index(h1_header_prefix) + 2].isspace()
            and len(current_line) > 2
        ):
            return (
                current_line[current_line.index(h1_header_prefix) :].strip("#").strip()
            )

    elif "=" in current_line:
        # Check if we're inspecting an empty or undefined lines.
        if not prev_line:
            return ""

        # Check if the current line only has equal sign divider.
        if current_line.count("=") == len(current_line.strip()):
            # Update header to the previous line.
            return prev_line.strip()

    return ""


def _extract_header_from_markdown(mdfile: Iterable[str]) -> str:
    """For a given markdown file, extract its header line.

    Args:
        mdfile: iterator to the markdown file.

    Returns:
        A string for header or empty string if header is not found.
    """
    prev_line = ""

    for line in mdfile:
        # Ignore licenses and other non-headers prior to the header.
        # If we've found the header, return the header.
        header = _parse_markdown_header(line, prev_line)
        if header != "":
            return header

        prev_line = line

    return ""


def _remove_license(mdfile_path: str) -> None:
    """Removes any licenses in markdown files.

    Args:
        mdfile_path (str): The path to the markdown file.
    """

    comment_tag_begin = "<!--"
    comment_tag_end = "-->"

    with open(mdfile_path) as mdfile:
        file_content = mdfile.read()

    # Find the first occurrence of comment tags.
    begin_index = file_content.find(comment_tag_begin)
    end_index = file_content.find(comment_tag_end)

    # Check whether the HTML comment is a license - they should be at the top of
    # the file, and if any content prior to the license is visible other than
    # whitespace we assume it's not a license comment.
    if (pre_comment := file_content[:begin_index]) and not pre_comment.isspace():
        return

    # Strip the license.
    file_content = file_content[end_index + len(comment_tag_end) :]

    # Reset file position to the beginning to write
    with open(mdfile_path, "w") as mdfile:
        mdfile.write(file_content)


def _highlight_md_codeblocks(mdfile_path: str) -> None:
    """Adds syntax highlighting to code blocks for a given markdown file.

    Args:
        mdfile_path (str): The path to the markdown file.
    """
    fence = "```"
    fence_with_python = "```python"
    new_lines = []

    with open(mdfile_path) as mdfile:
        file_content = mdfile.read()
        # If there is an odd number of code block annotations, do not syntax
        # highlight.
        if file_content.count(fence) % 2 != 0:
            print(
                f"{mdfile.name} contains wrong format of code blocks. Skipping syntax highlighting."
            )
            return
        # Retrieve code block positions to replace
        codeblocks = [[m.start(), m.end()] for m in re.finditer(fence, file_content)]

        # This is equivalent to grabbing every odd index item.
        codeblocks = codeblocks[::2]
        # Used to store code blocks that come without language indicators.
        blocks_without_indicators = []

        # Check if the fence comes without a language indicator. If so, include
        # this to a list to render.
        for start, end in codeblocks:
            if file_content[end] == "\n":
                blocks_without_indicators.append([start, end])

        # Stitch content that does not need to be parsed, and replace with
        # `fence_with_python` for parsed portions.
        prev_start = prev_end = 0
        for start, end in blocks_without_indicators:
            new_lines.append(file_content[prev_end:start])
            new_lines.append(fence_with_python)
            prev_start, prev_end = start, end

        # Include rest of the content.
        new_lines.append(file_content[prev_end:])

    # Overwrite with newly parsed content.
    with open(mdfile_path, "w") as mdfile:
        new_content = "".join(new_lines)
        mdfile.write(new_content)


def _clean_image_links(mdfile_path: str) -> None:
    """Cleans extra whitespace that breaks image links in index.html file.

    Args:
        mdfile_path (str): The path to the markdown file.
    """
    image_link_pattern = r"\[\s*!\[image\]\(.*\)\s*\]\(.*\)"
    new_lines = []
    with open(mdfile_path) as mdfile:
        file_content = mdfile.read()

        prev_start = prev_end = 0

        for matched_obj in re.finditer(image_link_pattern, file_content):
            start = matched_obj.start()
            end = matched_obj.end()
            matched_str = file_content[start:end]
            # Clean up all whitespaces for the image link.
            clean_str = "".join(matched_str.split())

            new_lines.append(file_content[prev_end:start])
            new_lines.append(clean_str)
            prev_start, prev_end = start, end

        new_lines.append(file_content[prev_end:])

    with open(mdfile_path, "w") as mdfile:
        new_content = "".join(new_lines)
        mdfile.write(new_content)


def _prepend_markdown_header(filename: str, mdfile: Iterable[str]) -> None:
    """Prepends the filename as a Markdown header.

    Args:
        filename: the name of the markdown file to prepend.
        mdfile: iterator to the markdown file that is both readable
          and writable.
    """
    file_content = f"# {filename}\n\n" + mdfile.read()
    # Reset file position to the beginning to write
    mdfile.seek(0)
    mdfile.write(file_content)


def _merge_markdown_content(
    *,
    base_file: str,
    additional_content_file: str,
    prepend_additional_content: bool = False,
) -> None:
    """Merges the Markdown file contents.

    The additional_content_file's contents get appended to the base_file's
    content.

    Args:
        base_file: The base content to append content to.
        additional_content_file: The additional content to be appended to
            the base file.
        prepend_additional_content: Optional. If specified, puts the additional
            content before the base file content.
    """
    try:
        with open(base_file, "r+") as base, open(
            additional_content_file, "r"
        ) as additional_content:
            file_content = (
                f"{additional_content.read()}\n{base.read()}"
                if prepend_additional_content
                else f"{base.read()}\n{additional_content.read()}"
            )
            base.seek(0)
            base.write(file_content)
    except OSError:
        print("Could not open the given files.")
        return


def move_markdown_pages(
    app: sphinx.application,
    outdir: Path,
    cwd: list[str] | None = [],
) -> None:
    """Moves markdown pages to be added to the generated reference documentation.

    Markdown pages may be hand written or auto generated. They're processed
    through a third party library to process markdown, then does further
    processing here then added to the top level of the TOC.

    Args:
        app: Sphinx application.
        outdir: The output directory to move markdown pages to.
    """
    # Use this to ignore markdown files that are unnecessary.
    files_to_ignore = [
        "index.md",  # use readme.md instead
        "reference.md",  # Reference docs overlap with Overview. Will try and incorporate this in later.
        # See https://github.com/googleapis/sphinx-docfx-yaml/issues/106.
        "overview_content.md",  # Content will be merged into summary_overview page.
    ]

    # Use this to move the page but do not add them in the TOC.
    ignore_in_toc = [
        "summary_overview.md",  # Already included in the TOC with other
        # summary pages.
    ]

    files_to_rename = {
        "readme.md": "index.md",
    }

    base_markdown_dir = Path(app.builder.outdir).parent / "markdown"

    markdown_dir = base_markdown_dir.joinpath(*cwd) if cwd else base_markdown_dir

    if not markdown_dir.exists():
        print("There's no markdown file to move.")
        return

    # Used to keep track of the index page entry to insert later.
    index_page_entry = None

    markdown_file_names = {mdfile.name.lower() for mdfile in markdown_dir.iterdir()}

    # If there is an index.md and no readme.md, use the index.md. Otherwise, we ignore the index.md.
    if "index.md" in markdown_file_names and "readme.md" not in markdown_file_names:
        files_to_ignore.remove("index.md")

    if (
        "summary_overview.md" in markdown_file_names
        and "overview_content.md" in markdown_file_names
    ):
        # Keep the summary_overview file, prepend the additioanl content.
        _merge_markdown_content(
            base_file=markdown_dir / "summary_overview.md",
            additional_content_file=markdown_dir / "overview_content.md",
            prepend_additional_content=True,
        )

    # For each file, if it is a markdown file move to the top level pages.
    for mdfile in markdown_dir.iterdir():
        if mdfile.is_dir():
            cwd.append(mdfile.name)
            move_markdown_pages(app, outdir, cwd)
            # Restore the original cwd after finish working on the directory.
            cwd.pop()

        if not mdfile.is_file() or mdfile.name.lower() in files_to_ignore:
            continue

        mdfile_name = ""

        _remove_license(mdfile)

        # Extract the header name for TOC.
        with open(mdfile) as mdfile_iterator:
            name = _extract_header_from_markdown(mdfile_iterator)

        if not name:
            with open(mdfile, "r+") as mdfile_iterator:
                mdfile_name = (
                    mdfile_iterator.name.split("/")[-1].split(".")[0].capitalize()
                )

                print(
                    f"Could not find a title for {mdfile_iterator.name}. Using {mdfile_name} as the title instead."
                )
                name = mdfile_name

                _prepend_markdown_header(name, mdfile_iterator)

        mdfile_name_to_use = mdfile.name.lower()
        if mdfile_name_to_use in files_to_rename:
            mdfile_name_to_use = files_to_rename[mdfile_name_to_use]

        if cwd and mdfile_name_to_use == "index.md":
            mdfile_name_to_use = f"{'_'.join(cwd)}_{mdfile_name_to_use}"

        mdfile_outdir = f"{outdir}/{mdfile_name_to_use}"

        shutil.copy(mdfile, mdfile_outdir)

        _highlight_md_codeblocks(mdfile_outdir)
        _clean_image_links(mdfile_outdir)

        if mdfile.name.lower() in ignore_in_toc:
            continue

        app.env.moved_markdown_pages.add(mdfile_name_to_use)

        if not cwd:
            # Use Overview as the name for top-level index file.
            if "index.md" in mdfile_name_to_use:
                # Save the index page entry.
                index_page_entry = {
                    "name": "Overview",
                    "href": "index.md",
                }
                continue

            # Use '/' to reserve for top level pages.
            app.env.markdown_pages["/"].append(
                {
                    "name": name,
                    "href": mdfile_name_to_use,
                }
            )
            continue

        # Add the file to the TOC later.
        app.env.markdown_pages[cwd[-1]].append(
            {
                "name": name,
                "href": mdfile_name_to_use,
            }
        )

    if app.env.markdown_pages.get("/"):
        # Sort the top level pages. Other pages will be sorted when they're
        # added to package level files accordingly.
        app.env.markdown_pages["/"] = sorted(
            app.env.markdown_pages["/"],
            key=lambda entry: entry["href"],
        )

    if index_page_entry is None:
        return

    # Place the Overview page at the top of the list.
    app.env.markdown_pages["/"].insert(
        0,
        index_page_entry,
    )


def remove_unused_pages(
    added_pages: MutableSet[str],
    all_pages: MutableSet[str],
    outdir: Path,
) -> None:
    """Removes unused markdown pages after merging the table of contents.

    Pages may be generated as part of generating the document. API pages
    are needed and may be generated as part of Sphinx config, but if not
    used they will be identified and removed.

    Args:
        added_pages: markdown pages that have been added to the merged
            table of contents.
        all_pages: set of all markdown pages generated.
        outdir: output directory containing the markdown pages.
    """

    pages_to_remove = set(page for page in all_pages if page not in added_pages)

    for page in pages_to_remove:
        try:
            os.remove(f"{outdir}/{page}")
        except FileNotFoundError:
            # This shouldn't happen, but in case we fail, ignore the failure
            # and continue deleting other files.
            print(f"Could not delete {page}.")


def run_sphinx_markdown(app: sphinx.application) -> None:
    """Runs sphinx-build with Markdown builder in the plugin.

    Args:
        app (sphinx.application): The sphinx application.
    """
    cwd = os.getcwd()
    relative_srcdir = app.srcdir.removeprefix(f"{cwd}/")
    relative_outdir = app.outdir.removeprefix(f"{cwd}/").removesuffix("/html")
    # Skip running sphinx-build for Markdown for some unit tests.
    # Not required other than to output DocFX YAML.
    if "docs" in cwd:
        return

    return shell.run(
        [
            "sphinx-build",
            "-M",
            "markdown",
            relative_srcdir,
            relative_outdir,
        ],
        hide_output=False,
    )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
**DEPRECATED**

This was an approach that pulls domain data from doctree.

This approach will work for non-autodoc domains.
However,
the approach used in this extension is easiest for autodoc.

In order to build the approach going forward,
a bit more work needs to be done to pull docstrings in raw format.
"""

from docutils import nodes
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.util.console import bold, red
from sphinx.util.docfields import _is_single_paragraph

from .utils import _get_desc_data

TITLE_MAP = {
    "Returns": "return",
    "Return type": "return_type",
    "Raises": "raises",
    "Parameters": "params",
}


def doctree_resolved(app: Sphinx, doctree: nodes.document, docname: str) -> None:
    """
    Render out the YAML from the Sphinx domain objects

    Args:
        app (Sphinx): The sphinx application.
        doctree (nodes.document): The doctree.
        docname (str): The name of the document.
    """
    yaml_data, yaml_modules = extract_yaml(app, doctree, ignore_patterns=None)
    extract_info_lists(app, doctree)
    # if yaml_data:
    #     app.env.docfx_yaml_data[docname] = yaml_data
    # if yaml_modules:
    #     for module in yaml_modules:
    #         app.env.docfx_yaml_modules[module] = yaml_modules[module]


def _get_full_data(node: nodes.Node) -> dict:
    """Extracts data from a field list node.

    Args:
        node (nodes.Node): The field list node.

    Returns:
        dict: The extracted data.
    """
    data = {}
    parent_desc = node.parent.parent
    name, uid = _get_desc_data(parent_desc)
    if not name:
        return

    for field in node:
        fieldname, fieldbody = field
        try:
            # split into field type and argument
            fieldtype, _ = fieldname.astext().split(None, 1)
        except ValueError:
            # maybe an argument-less field type?
            fieldtype = fieldname.astext()

        # collect the content, trying not to keep unnecessary paragraphs
        if _is_single_paragraph(fieldbody):
            content = fieldbody.children[0].children
        else:
            content = fieldbody.children

        data.setdefault(uid, {})

        if fieldtype == "Returns":
            for child in content:
                ret_data = child.astext()
                data[uid].setdefault(fieldtype, []).append(ret_data)

        if fieldtype == "Raises":
            for child in content:
                ret_data = child.astext()
                data[uid].setdefault(fieldtype, []).append(ret_data)
    return data


def extract_info_lists(app: Sphinx, doctree: nodes.document) -> None:
    """Extracts info lists from the doctree.

    Args:
        app (Sphinx): The sphinx application.
        doctree (nodes.document): The doctree.
    """
    data = []

    for node in doctree.traverse(nodes.field_list):
        data.append(_get_full_data(node))

    print(data)


def extract_yaml(
    app: Sphinx, doctree: nodes.document, ignore_patterns: list[str]
) -> tuple[list[dict], dict]:
    """Iterate over all Python domain objects and output YAML.

    Args:
        app (Sphinx): The sphinx application.
        doctree (nodes.document): The doctree.
        ignore_patterns (list[str]): A list of patterns to ignore.

    Returns:
        tuple[list[dict], dict]: A tuple containing the list of items and
            a dictionary of modules.
    """
    items = []
    modules = {}

    for desc_node in doctree.traverse(addnodes.desc):
        if desc_node.attributes["domain"] != "py":
            app.info(
                bold("[docfx_yaml] ")
                + red("Skipping Domain Object (%s)" % desc_node.attributes["domain"])
            )
            continue

        module = desc_node[0].attributes["module"]

        if not module:
            app.info(bold("[docfx_yaml] ") + red("Skipping object with no module"))
            continue

        if module not in modules:
            modules[module] = [
                {
                    "module": str(module),
                    "uid": str(module),
                    "type": "Namespace",
                    "type": "module",
                    "name": str(module),
                    "children": [],
                }
            ]

        _type = desc_node.attributes["objtype"]
        full_name = desc_node[0].attributes["fullname"]
        try:
            uid = desc_node[0].attributes["ids"][0]
        except Exception:
            uid = f"{module}.{full_name}"
            print("Non-standard id: %s" % uid)
        name = desc_node[0].attributes["names"][0]
        source = desc_node[0].source
        try:
            summary = desc_node[1][0].astext()
        except (KeyError, IndexError):
            summary = ""

        try:
            args = [arg.strip() for arg in desc_node[0][3].astext().split(",")]
        except Exception:
            args = []

        if args:
            full_name += f"({', '.join(args)})"

        datam = {
            "module": str(module),
            "uid": uid,
            "type": _type,
            "name": name,
            "fullName": full_name,
            "summary": summary,
            "rst_source": source,
        }

        if _type == "method":
            datam["class"] = ".".join(name.split(".")[:-1])
        if _type == "class":
            datam["children"] = []

        # insert_children(_type, datam, modules)
        items.append(datam)

        modules[module].append(datam)

    return (items, modules)



"""Tests for various inspect related utils."""

import inspect
import unittest
from typing import Any, Optional, Union

from parameterized import parameterized

from docfx_yaml import extension


class TestGenerate(unittest.TestCase):
    types_to_test = [
        [
            # Test for a simple type without __args__.
            str,
            "str",
        ],
        [
            # Test for a more complex type without forward reference.
            list[str],
            "list[str]",
        ],
        [
            # Test for imported type, without forward reference.
            dict[str, Any],
            "dict[str, typing.Any]",
        ],
        [
            # Test for forward reference.
            Optional["ForwardClass"],
            "typing.Optional[ForwardClass]",
        ],
        [
            # Test for multiple forward references.
            Union["ForwardClass", "ForwardClass2"],
            "typing.Union[ForwardClass, ForwardClass2]",
        ],
    ]

    @parameterized.expand(types_to_test)
    def test_extracts_annotations(self, type_to_test, expected_type_name):
        """Extracts annotations from test method, compares to expected name."""

        def test_method(name: type_to_test):
            pass

        annotations = inspect.getfullargspec(test_method).annotations
        annotation_to_use = annotations["name"]

        extracted_annotation_name = extension._extract_type_name(annotation_to_use)

        self.assertEqual(extracted_annotation_name, expected_type_name)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import namedtuple
from inspect import signature

from docutils import nodes
from docutils.io import StringOutput
from docutils.utils import new_document
from sphinx.application import Sphinx

from .writer import MarkdownWriter as Writer


def slugify(value: str) -> str:
    """Converts to lowercase, removes non-word characters.

    Converts non-word characters (alphanumerics and
    underscores) and converts spaces to hyphens. Also strips leading and
    trailing whitespace.

    Args:
        value (str): The string to convert.

    Returns:
        str: The converted string.
    """
    value = re.sub(r"[^\w\s-]", "", value).strip()
    return re.sub(r"[-\s]+", "-", value)


def transform_string(app: Sphinx, string: str) -> str:
    """Transforms a string by parsing it and then writing it out.

    This is useful for resolving references in the string.

    Args:
        app (Sphinx): The sphinx application.
        string (str): The string to transform.

    Returns:
        str: The transformed string.
    """
    ret = []
    for para in string.split("\n\n"):
        tmp = nodes.paragraph(para, para)
        ret.append(transform_node(app, tmp))
    return "\n\n".join(ret)


def transform_node(app: Sphinx, node: nodes.Node) -> str:
    """Transforms a docutils node to a string.

    Args:
        app (Sphinx): The sphinx application.
        node (nodes.Node): The node to transform.

    Returns:
        str: The transformed node as a string.
    """
    destination = StringOutput(encoding="utf-8")
    doc = new_document(b"<partial node>")
    doc.append(node)

    # Resolve refs
    doc["docname"] = "inmemory"
    app.env.resolve_references(doctree=doc, fromdocname="inmemory", builder=app.builder)

    writer = Writer(app.builder)
    writer.write(doc, destination)
    return destination.destination.decode("utf-8")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import setuptools

name = "gcp-sphinx-docfx-yaml"
description = "Sphinx Python Domain to DocFX YAML Generator"
version = "3.2.5"
dependencies = [
    "black",
    "gcp-docuploader",
    "PyYAML",
    "recommonmark",
    "sphinxcontrib-applehelp==1.0.4",
    "sphinxcontrib-devhelp==1.0.2",
    "sphinxcontrib-htmlhelp==2.0.1",
    "sphinxcontrib-qthelp==1.0.3",
    "sphinxcontrib-serializinghtml==1.1.5",
    "sphinx==4.5.0",
    "sphinx-markdown-builder==0.5.5",
    "sphinxcontrib.napoleon",
    "unidecode",
    "wheel>=0.24.0",
]

packages = setuptools.find_packages(".", exclude=["tests"])

extra_setup = dict(
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mock"],
)

setuptools.setup(
    name=name,
    version=version,
    description=description,
    author="Google LLC",
    author_email="dandhlee@google.com",
    license="Apache 2.0",
    url=(
        "https://github.com/googleapis/google-cloud-python/tree/main/"
        "packages/gcp-sphinx-docfx-yaml"
    ),
    package_dir={"": "."},
    packages=packages,
    install_requires=dependencies,
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    zip_safe=False,
    **extra_setup,
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
This module is used to add extra supported directives to sphinx.
"""

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives.admonitions import BaseAdmonition

from .nodes import remarks


class RemarksDirective(BaseAdmonition):
    """A directive to handle 'remarks' blocks."""

    node_class = remarks


class TodoDirective(Directive):
    """A directive to ignore 'todo' blocks."""

    # Enable content in the directive
    has_content = True

    def run(self) -> list[nodes.Node]:
        """This directive is ignored, so return empty list."""
        return_nodes = []

        return return_nodes



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from functools import partial
from typing import Any

import sphinx.application
from docutils import nodes
from sphinx import addnodes, directives
from sphinx.addnodes import desc, desc_signature
from sphinx.util import docfields
from sphinx.util.docfields import _is_single_paragraph

from .nodes import remarks
from .utils import transform_node as _transform_node

TYPE_SEP_PATTERN = r"(\[|\]|, |\(|\))"


def _get_desc_data(node: nodes.Node) -> tuple[str | None, str | None]:
    """Gets the description data from a node.
    Args:
        node (nodes.Node): The node to get the description data from.
    Returns:
        tuple[str | None, str | None]: A tuple containing the full name
            and uid of the node.
    """
    assert node.tagname == "desc"
    if node.attributes["domain"] != "py":
        print("Skipping Domain Object (%s)" % node.attributes["domain"])
        return None, None

    try:
        module = node[0].attributes["module"]
        full_name = node[0].attributes["fullname"].split(".")[-1]
    except KeyError as e:
        print(
            "[docfx_yaml] There maybe some syntax error in docstring near: "
            + node.astext()
        )
        raise e

    try:
        uid = node[0].attributes["ids"][0]
    except Exception:
        uid = f"{module}.{full_name}"
        print("Non-standard id: %s" % uid)
    return full_name, uid


def _is_desc_of_enum_class(node: nodes.Node) -> bool:
    """Checks if the node is a description of an enum class.
    Args:
        node (nodes.Node): The node to check.
    Returns:
        bool: True if the node is a description of an enum class,
            False otherwise.
    """
    assert node.tagname == "desc_content"
    if (
        node[0]
        and node[0].tagname == "paragraph"
        and node[0].astext() == "Bases: enum.Enum"
    ):
        return True

    return False


def _hacked_transform(typemap: dict, node: nodes.Node) -> tuple[list, dict]:
    """
    Taken from docfields.py from sphinx.

    This does all the steps around gathering data,
    but doesn't actually do the node transformations.
    """
    entries = []
    groupindices = {}
    types = {}

    # step 1: traverse all fields and collect field types and content
    for field in node:
        fieldname, fieldbody = field
        try:
            # split into field type and argument
            fieldtype, fieldarg = fieldname.astext().split(None, 1)
        except ValueError:
            # maybe an argument-less field type?
            fieldtype, fieldarg = fieldname.astext(), ""
        typedesc, is_typefield = typemap.get(fieldtype, (None, None))

        # sort out unknown fields
        if typedesc is None or typedesc.has_arg != bool(fieldarg):
            # either the field name is unknown, or the argument doesn't
            # match the spec; capitalize field name and be done with it
            new_fieldname = fieldtype[0:1].upper() + fieldtype[1:]
            if fieldarg:
                new_fieldname += " " + fieldarg
            fieldname[0] = nodes.Text(new_fieldname)
            entries.append(field)
            continue

        typename = typedesc.name

        # collect the content, trying not to keep unnecessary paragraphs
        if _is_single_paragraph(fieldbody):
            content = fieldbody.children[0].children
        else:
            content = fieldbody.children

        # if the field specifies a type, put it in the types collection
        if is_typefield:
            # filter out only inline nodes; others will result in invalid
            # markup being written out
            content = [
                n
                for n in content
                if isinstance(n, nodes.Inline) or isinstance(n, nodes.Text)
            ]
            if content:
                types.setdefault(typename, {})[fieldarg] = content
            continue

        # also support syntax like ``:param type name:``
        if typedesc.is_typed:
            try:
                argtype, argname = fieldarg.split(None, 1)
            except ValueError:
                pass
            else:
                types.setdefault(typename, {})[argname] = [nodes.Text(argtype)]
                fieldarg = argname

        translatable_content = nodes.inline(fieldbody.rawsource, translatable=True)
        translatable_content.source = fieldbody.parent.source
        translatable_content.line = fieldbody.parent.line
        translatable_content += content

        # grouped entries need to be collected in one entry, while others
        # get one entry per field
        if typedesc.is_grouped:
            if typename in groupindices:
                group = entries[groupindices[typename]]
            else:
                groupindices[typename] = len(entries)
                group = [typedesc, []]
                entries.append(group)
            entry = typedesc.make_entry(fieldarg, [translatable_content])
            group[1].append(entry)
        else:
            entry = typedesc.make_entry(fieldarg, [translatable_content])
            entries.append([typedesc, entry])

    return (entries, types)


def patch_docfields(app: sphinx.application.Sphinx) -> None:
    """
    Grab syntax data from the Sphinx info fields.

    This is done by monkeypatching into the DocFieldTransformer,
    which is what Sphinx uses to transform the docutils ``nodes.field``
    into the sphinx ``docfields.Field`` objects.

    See usage in Sphinx
    `here <https://github.com/sphinx-doc/sphinx/blob/master/sphinx/directives/__init__.py#L180>`_.

    This also performs the RST doctree to Markdown transformation on the content,
    using the :class:`docfx_yaml.writers.MarkdownWriter`.
    """

    transform_node = partial(_transform_node, app)

    def get_data_structure(
        entries: list, types: dict, field_object: nodes.Node
    ) -> dict:
        """
        Get a proper docfx YAML data structure from the entries & types
        """

        data = {
            "parameters": [],
            "variables": [],
            "exceptions": [],
            "return": {},
            "references": [],
        }

        def make_param(_id, _description, _type=None, _required=None):
            ret = {"id": _id, "description": _description.strip(" \n\r\t")}
            if _type:
                ret["type"] = _type

            if _required is not None:
                ret["isRequired"] = _required

            return ret

        def transform_para(para_field):
            if isinstance(para_field, addnodes.pending_xref):
                return transform_node(para_field)
            else:
                return para_field.astext()

        def resolve_type(data_type):
            # Remove @ ~ and \n for cross reference in parameter/return value type to apply to docfx correctly
            data_type = re.sub("[@~\n]", "", data_type)

            # Add references for docfx to resolve ref if type contains TYPE_SEP_PATTERN
            _spec_list = []
            _spec_fullnames = re.split(TYPE_SEP_PATTERN, data_type)

            _added_reference = {}
            if len(_spec_fullnames) > 1:
                _added_reference_name = ""
                for _spec_fullname in _spec_fullnames:
                    if _spec_fullname != "":
                        _spec = {}
                        _spec["name"] = _spec_fullname.split(".")[-1]
                        _spec["fullName"] = _spec_fullname
                        if re.match(TYPE_SEP_PATTERN, _spec_fullname) is None:
                            _spec["uid"] = _spec_fullname
                        _spec_list.append(_spec)
                        _added_reference_name += _spec["name"]

                _added_reference = {
                    "uid": data_type,
                    "name": _added_reference_name,
                    "fullName": data_type,
                    "spec.python": _spec_list,
                }

            return data_type, _added_reference

        def extract_exception_desc(field_object):
            ret = []
            if len(field_object) > 0:
                for field in field_object:
                    if (
                        "field_name" == field[0].tagname
                        and field[0].astext() == "Raises"
                    ):
                        assert field[1].tagname == "field_body"
                        field_body = field[1]

                        children = [
                            n for n in field_body if not isinstance(n, nodes.Invisible)
                        ]

                        for child in children:
                            if isinstance(child, nodes.paragraph):
                                pending_xref_index = child.first_child_matching_class(
                                    addnodes.pending_xref
                                )
                                if pending_xref_index is not None:
                                    pending_xref = child[pending_xref_index]
                                    raise_type_index = (
                                        pending_xref.first_child_matching_class(
                                            nodes.literal
                                        )
                                    )
                                    if raise_type_index is not None:
                                        raise_type = pending_xref[raise_type_index]
                                        ret.append(
                                            {
                                                "type": pending_xref["reftarget"],
                                                "desc": raise_type.astext(),
                                            }
                                        )

            return ret

        for entry in entries:
            if isinstance(entry, nodes.field):
                # pass-through old field
                pass
            else:
                fieldtype, content = entry
                fieldtypes = types.get(fieldtype.name, {})
                if fieldtype.name == "exceptions":
                    for _type, _description in content:
                        data["exceptions"].append(
                            {
                                "type": _type,
                                "description": transform_node(_description[0]).strip(
                                    " \n\r\t"
                                ),
                            }
                        )
                if fieldtype.name == "returntype":
                    for returntype_node in content[1]:
                        returntype_ret = transform_node(returntype_node)
                        if returntype_ret:
                            # Support or in returntype
                            for returntype in re.split("[ \n]or[ \n]", returntype_ret):
                                returntype, _added_reference = resolve_type(returntype)
                                if _added_reference:
                                    if len(data["references"]) == 0:
                                        data["references"].append(_added_reference)
                                    elif any(
                                        r["uid"] != _added_reference["uid"]
                                        for r in data["references"]
                                    ):
                                        data["references"].append(_added_reference)

                                data["return"].setdefault("type", []).append(returntype)
                if fieldtype.name == "returnvalue":
                    returnvalue_ret = transform_node(content[1][0])
                    if returnvalue_ret:
                        data["return"]["description"] = returnvalue_ret.strip(" \n\r\t")
                if fieldtype.name in ["parameter", "variable", "keyword"]:
                    for field, node_list in content:
                        _id = field
                        _description = transform_node(node_list[0])
                        if field in fieldtypes:
                            _type = "".join(
                                transform_para(n) for n in fieldtypes[field]
                            )
                        else:
                            _type = None

                        _para_types = []
                        if fieldtype.name == "parameter" or fieldtype.name == "keyword":
                            if _type:
                                # Support or in parameter type
                                for _s_type in re.split("[ \n]or[ \n]", _type):
                                    _s_type, _added_reference = resolve_type(_s_type)
                                    if _added_reference:
                                        if len(data["references"]) == 0:
                                            data["references"].append(_added_reference)
                                        elif any(
                                            r["uid"] != _added_reference["uid"]
                                            for r in data["references"]
                                        ):
                                            data["references"].append(_added_reference)

                                    _para_types.append(_s_type)

                            _data = make_param(
                                _id=_id,
                                _type=_para_types,
                                _description=_description,
                                _required=False
                                if fieldtype.name == "keyword"
                                else True,
                            )
                            data["parameters"].append(_data)

                        if fieldtype.name == "variable":
                            if _type:
                                # Support or in variable type
                                for _s_type in re.split("[ \n]or[ \n]", _type):
                                    _s_type, _added_reference = resolve_type(_s_type)
                                    if _added_reference:
                                        if len(data["references"]) == 0:
                                            data["references"].append(_added_reference)
                                        elif any(
                                            r["uid"] != _added_reference["uid"]
                                            for r in data["references"]
                                        ):
                                            data["references"].append(_added_reference)

                                    _para_types.append(_s_type)

                            _data = make_param(
                                _id=_id, _type=_para_types, _description=_description
                            )
                            data["variables"].append(_data)

                    ret_list = extract_exception_desc(field_object)
                    for ret in ret_list:
                        # only use type in exceptions
                        data.setdefault("exceptions", []).append({"type": ret["type"]})

        return data

    class PatchedDocFieldTransformer(docfields.DocFieldTransformer):
        @staticmethod
        def type_mapping(type_name: str) -> str:
            """Maps a type name to a docfx type name.
            Args:
                type_name (str): The type name to map.
            Returns:
                str: The mapped type name.
            """
            mapping = {
                "staticmethod": "method",
                "classmethod": "method",
                "exception": "class",
            }

            return mapping[type_name] if type_name in mapping else type_name

        def __init__(self, directive: Any):
            """Initializes the transformer.
            Args:
                directive (Any): The directive.
            """
            self.directive = directive
            super(PatchedDocFieldTransformer, self).__init__(directive)

        def transform_all(self, node: nodes.Node) -> None:
            """Transform all field list children of a node."""
            # don't traverse, only handle field lists that are immediate children
            summary = []
            data = {}
            name, uid = _get_desc_data(node.parent)
            for child in node:
                if isinstance(child, remarks):
                    remarks_string = transform_node(child)
                    data["remarks"] = remarks_string
                elif isinstance(child, addnodes.desc):
                    if child.get("desctype") == "attribute":
                        attribute_map = {}  # Used for detecting duplicated attributes in intermediate data and merge them

                        for item in child:
                            if isinstance(item, desc_signature) and any(
                                isinstance(n, addnodes.desc_annotation) for n in item
                            ):
                                # capture attributes data and cache it
                                data.setdefault("added_attribute", [])

                                item_ids = item.get("ids", [""])

                                if (
                                    len(item_ids) == 0
                                ):  # find a node with no 'ids' attribute
                                    curuid = (
                                        item.get("module", "")
                                        + "."
                                        + item.get("fullname", "")
                                    )
                                    # generate its uid by module and fullname
                                else:
                                    curuid = item_ids[0]

                                if len(curuid) > 0:
                                    parent = curuid[: curuid.rfind(".")]
                                    name = item.children[0].astext()

                                    if curuid in attribute_map:
                                        if (
                                            len(item_ids) == 0
                                        ):  # ensure the order of docstring attributes and real attributes is fixed
                                            attribute_map[curuid]["syntax"][
                                                "content"
                                            ] += " " + item.astext()
                                            # concat the description of duplicated nodes
                                        else:
                                            attribute_map[curuid]["syntax"][
                                                "content"
                                            ] = (
                                                item.astext()
                                                + " "
                                                + attribute_map[curuid]["syntax"][
                                                    "content"
                                                ]
                                            )
                                    else:
                                        if _is_desc_of_enum_class(node):
                                            addedData = {
                                                "uid": curuid,
                                                "id": name,
                                                "parent": parent,
                                                "langs": ["python"],
                                                "name": name,
                                                "fullName": curuid,
                                                "type": item.parent.get("desctype"),
                                                "module": item.get("module"),
                                                "syntax": {
                                                    "content": item.astext(),
                                                    "return": {"type": [parent]},
                                                },
                                            }
                                        else:
                                            addedData = {
                                                "uid": curuid,
                                                "class": parent,
                                                "langs": ["python"],
                                                "name": name,
                                                "fullName": curuid,
                                                "type": "attribute",
                                                "module": item.get("module"),
                                                "syntax": {"content": item.astext()},
                                            }

                                        attribute_map[curuid] = addedData
                                else:
                                    raise Exception(
                                        "ids of node: " + repr(item) + " is missing."
                                    )
                                    # no ids and no duplicate or uid can not be generated.
                        if "added_attribute" in data:
                            data["added_attribute"].extend(
                                attribute_map.values()
                            )  # Add attributes data to a temp list

                    # Don't recurse into child nodes
                    continue
                elif isinstance(child, nodes.field_list):
                    (entries, types) = _hacked_transform(self.typemap, child)
                    _data = get_data_structure(entries, types, child)
                    data.update(_data)
                elif isinstance(child, addnodes.seealso):
                    data["seealso"] = transform_node(child)
                elif (
                    isinstance(child, nodes.admonition)
                    and "Example" in child[0].astext()
                ):
                    # Remove the admonition node
                    child_copy = child.deepcopy()
                    child_copy.pop(0)
                    data["example"] = transform_node(child_copy)
                else:
                    content = transform_node(child)

                    # skip 'Bases' in summary
                    if not content.startswith("Bases: "):
                        summary.append(content)

            if "desctype" in node.parent and node.parent["desctype"] == "class":
                data.pop(
                    "exceptions", ""
                )  # Make sure class doesn't have 'exceptions' field.

            if summary:
                data["summary"] = "\n".join(summary)
            # Don't include empty data
            for key, val in data.copy().items():
                if not val:
                    del data[key]
            data["type"] = (
                PatchedDocFieldTransformer.type_mapping(node.parent["desctype"])
                if "desctype" in node.parent
                else "unknown"
            )
            self.directive.env.docfx_info_field_data[uid] = data
            super(PatchedDocFieldTransformer, self).transform_all(node)

    directives.DocFieldTransformer = PatchedDocFieldTransformer
